# This is a relay I run on a personal server. If it gets too expensive to
# run, I'll shut it down.
MAILBOX_RELAY = RENDEZVOUS_RELAY = "ws://relay.magic-wormhole.io:4000/v1"
TRANSIT_RELAY = "tcp:transit.magic-wormhole.io:4001"
